//********************************************************************
//*** ResourcePool/ResourcePoolEventObserver.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePoolEventObserver.cc,v 1.10 2009-03-02 14:16:44 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include <map>
#include <limits>
#include "util/Clone.hh"
#include "ResourcePool/ResourcePoolEventObserver.hh"
#include "ResourcePool/ResourcePoolStreamLogger.hh" // for it's static members

namespace fatalmind {

ResourcePoolEventObserver::ResourcePoolEventObserver() 
: logevents(Event.NONE)
, names()
, nextEventBit(16)
{
	names[Event.INCREMENT_POOL] = "INCREMENT_POOL";
	names[Event.DOWNSIZE_POOL]  = "DOWNSIZE_POOL";
	names[Event.FAIL_RESOURCE]  = "FAIL_RESOURCE";
	names[Event.GET_RESOURCE_FAIL]  = "GET_RESOURCE_FAIL";
	names[Event.EXECUTE_NFE]    = "EXECUTE_NFE";
	names[Event.EXECUTE_FAIL]   = "EXECUTE_FAIL";
	names[Event.TIME_HIRES]     = "TIME_HIRES";
	names[Event.FAIL]           = "FAIL";
	names[Event.ALL]            = "ALL";
	setDefaultEventMask();
}

void ResourcePoolEventObserver::setDefaultEventMask() {
	char* e = getenv("FATALMIND_DEFAULT_LOGEVENTS");
	if (e != NULL) {
		std::string estr(e);
		logevents = getEventMask(estr);
	}
}

void ResourcePoolEventObserver::event(event_t eventID,  const std::string& resourceID, const std::string& message) {
	// default implementation does nothing
};

const std::string& ResourcePoolEventObserver::getEventName(const event_t event) const {
	namemap_t::const_iterator h = names.find(event);	
	if (h == names.end()) {
		// unkown event, no namemapping available for that.
		return INVALID;
	} else {
		return h->second;
	}
}	

// The C++ Standard Library (Josuttis), Paragraph 6.6
template <class K, class V>
class value_equals {
	private:
		V value;
	public:
		value_equals(const V& v)
		: value(v)
		{
		}

		bool operator() (std::pair<const K, V> elem) {
			return elem.second == value;
		}
};

const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::getEvent(const std::string& value) const {
	namemap_t::const_iterator end = names.end();
	namemap_t::const_iterator hit = std::find_if(names.begin(), end, value_equals<event_t, std::string>(value));
	if (hit != end) {
		return hit->first;
	} else {
		return Event.NONE;
	}
}

const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::getEventMask(const std::string& str) const {
	std::string delimiters = "| ";
	event_t rv = Event.NONE;
	// Skip delimiters at beginning.
	std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	// Find first "non-delimiter".
	std::string::size_type pos     = str.find_first_of(delimiters, lastPos);

	while (std::string::npos != pos || std::string::npos != lastPos) {
		// Found a token, "or" it to the mask
		rv |= getEvent(str.substr(lastPos, pos - lastPos));
		// Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
		// Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
	return rv;
}

const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::registerEvent(const std::string& value) {
	event_t event = getEvent(value);	
	if (event == Event.NONE) {
		if (nextEventBit < std::numeric_limits<event_t>::digits) {
			event = (1 << nextEventBit);
			nextEventBit++;
			names[event] = value;
		} else {
			return Event.NONE;
		}
	}
	return event;
}

Clone* ResourcePoolEventObserver::DoClone() const {
	return new ResourcePoolEventObserver(*this);
}

gc_ptr<EventReader> ResourcePoolEventObserver::getReader(const event_t event, const std::string& r)  {
	gc_ptr<EventReader> rv(new EventReader(*this, event, r)); // // TODO, bad referincing *this
	return rv;
}

EventReader::EventReader(ResourcePoolEventObserver& observ, ResourcePoolEventObserver::event_t e, std::string r) 
: super()
, observer(observ)
, event(e)
, resource(r)
{
}

/*
EventReader::EventReader(const EventReader& rho)
: super(rho)
, observer(rho.observer)
, event(rho.event)
, resource(rho.resource)
{
}
*/

EventReader::~EventReader() {
	if (observer.isEvent(event)) {
		observer.event(event, resource, str());
	}
}

EventReader& EventReader::operator<<(const std::string& val) {
	if (observer.isEvent(event)) {
		*((super*)this)<<val;
	}
	return *this;
}
EventReader& EventReader::operator<<(const char* val) {
	if (observer.isEvent(event)) {
		*((super*)this)<<val;
	}
	return *this;
}
EventReader& EventReader::operator<<(const char& val) {
	if (observer.isEvent(event)) {
		*((super*)this)<<val;
	}
	return *this;
}
EventReader& EventReader::operator<<(const int& val) {
	if (observer.isEvent(event)) {
		*((super*)this)<<val;
	}
	return *this;
}
EventReader& EventReader::operator<<(const unsigned int& val) {
	if (observer.isEvent(event)) {
		*((super*)this)<<val;
	}
	return *this;
}

/*
 * Static members
 */
const std::string ResourcePoolEventObserver::NoResourceID = "N/A";

const std::string ResourcePoolEventObserver::INVALID = "INVALID";
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::TIME_HIRES     = 1 << 0;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::INCREMENT_POOL = 1 << 1;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::DOWNSIZE_POOL  = 1 << 2;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::FAIL_RESOURCE  = 1 << 3;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::GET_RESOURCE_FAIL = 1 << 4;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::EXECUTE_NFE    = 1 << 5;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::EXECUTE_FAIL   = 1 << 6;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::FAIL           = FAIL_RESOURCE | GET_RESOURCE_FAIL  | EXECUTE_NFE | EXECUTE_FAIL;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::ALL            = std::numeric_limits<ResourcePoolEventObserver::event_t>::max() & ~TIME_HIRES;
const ResourcePoolEventObserver::event_t ResourcePoolEventObserver::Events::NONE           = 0;

ResourcePoolStreamLogger ResourcePoolStreamLogger::defaultLogger(std::clog);

} // namespace fatalmind

