//********************************************************************
//*** ResourcePool/oracle/Factory.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: Factory.hh,v 1.25 2009-03-02 14:16:45 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_ORACLE_FACTORY_HH
#define RESOURCEPOOL_ORACLE_FACTORY_HH

#ifndef RESOURCEPOOL_ORACLE_RESOURCEFWD_HH
#include "ResourceFwd.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_OCIENVHANDLEFWD_HH
#include "OCIEnvHandleFwd.hh"
#endif

#ifndef INCLUDED_GC_PTR_HH
#include "mm/gc_ptr.hh"
#endif

#include "ResourcePool/SQL/SQLFactoryInterface.hh"
#include "ResourcePool/ResourcePoolStreamLogger.hh"

#include <string>
#include <memory>

namespace fatalmind {

/**
 * A Factory for a oracle database connection (session).
 * Thats the place where you configure which database
 * user and password to use.
 *
 * The following event attributes are used:
 *
 * \b EXECUTE
 * \li \c BATCH \c ERROR - One or more of the statements in the batch had an error. The error details will be written in the next argument.
 * \li \c SLMDC - (not an error) A abbrevation for SelectListMetaDataCache. Means that the meta data for the select list (types and sizes of selectd columns) has been used from the cache so that an additional database round-trip could be avoided.
 * \li \c EXACT - (not an error) SQLSelectRow did successfully query for a single row (a database roundtrip was saved). 
 *
 * \b PREPARE
 * \li \c CACHED - (not an error) Means that the cursor was cached. e.g. the same statement was executed in the same session before, the prepare round-trip could be avoided.
 *
 * \b FETCH
 * \li \c NO \c DATA - a SQLSelectRow command did not return any data 
 * \li \c MORE \c DATA - a SQLSelectRow command did return more then one row (emulated ORA-01422 error)
 * 
 * \b COMMIT
 * \li \c IMPLICIT - The commit was performed in a single round trip with the previous command (database roundtrip saved).
 *
 * For all cases, ORA-xxxxx might appear to indicate an error.
 */
template<class TM /* a Threading model */
, template<class> class SLMDC = oracle::SelectListMetaDataCache 
 /* Alternatively you could use oracle::SelectListMetaDataNopCache */
> class oracleFactory : public SQL::SQLFactoryInterface {
		typedef SQL::SQLFactoryInterface super;
	public:
		/**
		 * \param service_name the TNS name of the database
		 *        Starting with Oracle 10g you can also use
		 *        the syntax [//]host[:port][/service name]
		 * \param user the user name
		 * \param pass the users password
		 * \param e    The Event Oberserver can be used to get notified about some
		 *             events like failure. See ResourcePoolEventObserver for details.
		 */
		oracleFactory(const std::string& service_name
			, const std::string& user
			, const std::string& pass
			, ResourcePoolEventObserver& e = ResourcePoolStreamLogger::defaultLogger
		);
		/**
		 * \param logon user/pass@service 
		 * \param e    The Event Oberserver can be used to get notified about some
		 *             events like failure. See ResourcePoolEventObserver for details.
		 */
		oracleFactory(const std::string& logon, ResourcePoolEventObserver& e = ResourcePoolStreamLogger::defaultLogger);
		oracleFactory(const oracleFactory<TM,SLMDC>&);
		~oracleFactory();

		oracleResource<TM>* create_resource() const;
	
		using super::databaseType_t;
		databaseType_t getDatabaseType() const {
			return OCI;
		};

	protected:
		virtual Clone* DoClone() const;

	private:
		std::string _service_name;
		std::string _user;
		std::string _pass;

		const unsigned int _mode;
		mutable std::auto_ptr<oracle::OCIEnvHandle> _env;
		gc_ptr<typename oracle::SelectListMetaDataCacheInterface<TM>::type, TM> _slmdc;
};

} // namespace fatalmind
#endif

